# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lxml import etree

import libvirt
from networking_mlnx._i18n import _LE, _LI, _LW
from oslo_log import log as logging

from networking_mlnx.eswitchd.common import constants
from networking_mlnx.eswitchd.db import device_db
from networking_mlnx.eswitchd.utils import pci_utils

LOG = logging.getLogger(__name__)


class ResourceManager(object):

    def __init__(self):
        self.pci_utils = pci_utils.pciUtils()
        self.device_db = device_db.DeviceDB()

    def add_fabric(self, fabric, pf, fabric_type):
        pci_id, hca_port, pf_mlx_dev = self._get_pf_details(pf)
        self.device_db.add_fabric(fabric, pf, pci_id, hca_port, fabric_type,
                                  pf_mlx_dev)
        vfs = self.discover_devices(pf)
        LOG.info(_LI("PF %(pf)s, vfs = %(vf)s"), {'pf': pf, 'vf': vfs})
        self.device_db.set_fabric_devices(fabric, vfs)

    def scan_attached_devices(self):
        devices = []
        vm_ids = {}
        conn = libvirt.openReadOnly('qemu:///system')
        domains = []
        self.macs_map = self._get_vfs_macs()
        domains_names = conn.listDefinedDomains()
        defined_domains = map(conn.lookupByName, domains_names)
        domains_ids = conn.listDomainsID()
        running_domains = map(conn.lookupByID, domains_ids)
        for domain in defined_domains:
            [state, maxmem, mem, ncpu, cputime] = domain.info()
            if state in (libvirt.VIR_DOMAIN_PAUSED,
                         libvirt.VIR_DOMAIN_SHUTDOWN,
                         libvirt.VIR_DOMAIN_SHUTOFF):
                domains.append(domain)
        domains += running_domains

        for domain in domains:
            raw_xml = domain.XMLDesc(0)
            tree = etree.XML(raw_xml)
            hostdevs = tree.xpath("devices/hostdev/source/address")
            vm_id = tree.find('uuid').text
            for dev in self._get_attached_hostdevs(hostdevs):
                devices.append(dev)
                vm_ids[dev[0]] = vm_id
        return devices, vm_ids

    def get_fabric_pf(self, fabric):
        return self.device_db.get_pf(fabric)

    def get_fabric_details(self, fabric):
        return self.device_db.get_fabric_details(fabric)

    def discover_devices(self, pf):
        return self.pci_utils.get_vfs_info(pf)

    def get_fabric_for_dev(self, dev):
        return self.device_db.get_dev_fabric(dev)

    def _get_vfs_macs(self):
        macs_map = {}
        fabrics = self.device_db.device_db.keys()
        for fabric in fabrics:
            fabric_details = self.device_db.get_fabric_details(fabric)
            try:
                macs_map[fabric] = \
                    self.pci_utils.get_vfs_macs_ib(fabric_details)
            except Exception:
                LOG.exception(_LE("Failed to get vfs macs for fabric %s "),
                              fabric)
                continue
        return macs_map

    def _get_attached_hostdevs(self, hostdevs):
        devs = []
        for hostdev in hostdevs:
            dev = self.pci_utils.get_device_address(hostdev)
            fabric = self.get_fabric_for_dev(dev)
            if fabric:
                fabric_details = self.get_fabric_details(fabric)
                if fabric_details['pf_device_type'] == \
                   constants.CX3_VF_DEVICE_TYPE:
                    hca_port = fabric_details['hca_port']
                    pf_mlx_dev = fabric_details['pf_mlx_dev']
                    vf_index = self.pci_utils.get_guid_index(pf_mlx_dev, dev,
                                                             hca_port)
                elif fabric_details['pf_device_type'] == \
                    constants.CX4_VF_DEVICE_TYPE:
                    vf_index = fabric_details['vfs'][dev]['vf_num']
                try:
                    mac = self.macs_map[fabric][str(vf_index)]
                    devs.append((dev, mac, fabric))
                except KeyError:
                    LOG.warning(_LW("Failed to retrieve Hostdev MAC"
                                    "for dev %s"), dev)
            else:
                LOG.info(_LI("No Fabric defined for device %s"), hostdev)
        return devs

    def _get_pf_details(self, pf):
        hca_port = self.pci_utils.get_eth_port(pf)
        pci_id = self.pci_utils.get_pf_pci(pf)
        pf_pci_id = self.pci_utils.get_pf_pci(pf, 'normal')
        pf_mlx_dev = self.pci_utils.get_pf_mlx_dev(pf_pci_id)
        return (pci_id, hca_port, pf_mlx_dev)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import sys

from networking_mlnx._i18n import _LE, _LI, _LW
from oslo_log import log as logging

from networking_mlnx.eswitchd.common import constants
from networking_mlnx.eswitchd.common import exceptions
from networking_mlnx.eswitchd.db import eswitch_db
from networking_mlnx.eswitchd.resource_mngr import ResourceManager
from networking_mlnx.eswitchd.utils.command_utils import execute
from networking_mlnx.eswitchd.utils import pci_utils


LOG = logging.getLogger(__name__)

INVALID_PKEY = 'none'
DEFAULT_PKEY_IDX = '0'
PARTIAL_PKEY_IDX = '1'
BASE_PKEY = '0x8000'
DEFAULT_MASK = 0x7fff
DEFAULT_PKEY = '0xffff'


class eSwitchHandler(object):

    def __init__(self, fabrics=None):
        self.eswitches = {}
        self.pci_utils = pci_utils.pciUtils()
        self.rm = ResourceManager()
        self.devices = set()
        if fabrics:
            self.add_fabrics(fabrics)

    def add_fabrics(self, fabrics):
        res_fabrics = []
        for fabric, pf in fabrics:
            fabric_type = None

            if pf in ('autoib', 'autoeth'):
                fabric_type = pf.strip('auto')
                pf = self.pci_utils.get_auto_pf(fabric_type)
            else:
                fabric_type = self.pci_utils.get_interface_type(pf)
                verify_vendor_pf = (
                    self.pci_utils.verify_vendor_pf(pf, constants.VENDOR))
                if (not verify_vendor_pf or
                        not self.pci_utils.is_sriov_pf(pf) or
                        not self.pci_utils.is_ifc_module(pf, fabric_type)):
                    LOG.error(_LE("PF %s must have Mellanox Vendor ID"
                              ",SR-IOV and driver module "
                              "enabled. Terminating!") % pf)
                    sys.exit(1)

            if fabric_type:
                self.eswitches[fabric] = eswitch_db.eSwitchDB()
                self._add_fabric(fabric, pf, fabric_type)
                res_fabrics.append((fabric, pf, fabric_type))
            else:
                LOG.info(_LI("No fabric type for PF:%s.Terminating!") % pf)
                sys.exit(1)
        self.sync_devices()

    def sync_devices(self):
        devices, vm_ids = self.rm.scan_attached_devices()
        added_devs = {}
        removed_devs = {}
        added_devs = set(devices) - self.devices
        removed_devs = self.devices - set(devices)
        self._treat_added_devices(added_devs, vm_ids)
        self._treat_removed_devices(removed_devs)
        self.devices = set(devices)

    def _add_fabric(self, fabric, pf, fabric_type):
        self.rm.add_fabric(fabric, pf, fabric_type)
        self._config_port_up(pf)
        fabric_details = self.rm.get_fabric_details(fabric)

        for vf in fabric_details['vfs']:
            self.eswitches[fabric].create_port(vf, constants.VIF_TYPE_HOSTDEV)

    def _treat_added_devices(self, devices, vm_ids):
        for device in devices:
            dev, mac, fabric = device
            if fabric:
                self.eswitches[fabric].attach_vnic(
                    port_name=dev, device_id=vm_ids[dev], vnic_mac=mac)
                if self.eswitches[fabric].vnic_exists(mac):
                    self.eswitches[fabric].plug_nic(port_name=dev)
            else:
                LOG.info(_LI("No Fabric defined for device %s"), dev)

    def _treat_removed_devices(self, devices):
        for dev, mac in devices:
            fabric = self.rm.get_fabric_for_dev(dev)
            if fabric:
                self.eswitches[fabric].detach_vnic(vnic_mac=mac)
            else:
                LOG.info(_LI("No Fabric defined for device %s"), dev)

    def set_fabric_mapping(self, fabric, interface):
        dev = self.rm.get_fabric_for_dev(interface)
        if not dev:
            fabrics = [(fabric, interface)]
            self.add_fabrics(fabrics)
            dev = interface
        return (fabric, interface)

    def get_vnics(self, fabrics):
        vnics = {}
        for fabric in fabrics:
            eswitch = self._get_vswitch_for_fabric(fabric)
            if eswitch:
                vnics_for_eswitch = eswitch.get_attached_vnics()
                vnics.update(vnics_for_eswitch)
            else:
                LOG.error(_LE("No eSwitch found for Fabric %s"), fabric)
                continue
        LOG.info(_LI("vnics are %s"), vnics)
        return vnics

    def create_port(self, fabric, vnic_type, device_id, vnic_mac, pci_slot):
        eswitch = self._get_vswitch_for_fabric(fabric)
        if eswitch:
            try:
                if eswitch.attach_vnic(
                        pci_slot, device_id, vnic_mac, pci_slot):
                    self._config_vf_mac_address(fabric, pci_slot, vnic_mac)
                else:
                    raise exceptions.MlxException('Failed to attach vnic')
            except (RuntimeError, exceptions.MlxException):
                LOG.error(_LE('Create port operation failed '))
                pci_slot = None
        else:
            LOG.error(_LE("No eSwitch found for Fabric %s"), fabric)

        return pci_slot

    def plug_nic(self, fabric, device_id, vnic_mac, pci_slot):
        eswitch = self._get_vswitch_for_fabric(fabric)
        if eswitch:
            eswitch.port_table[pci_slot]['vnic'] = vnic_mac
            eswitch.port_policy.update(
                {vnic_mac:
                    {'vlan': None,
                     'dev': pci_slot,
                     'device_id': device_id}})

            self._config_vf_mac_address(fabric, pci_slot, vnic_mac)
            eswitch.plug_nic(pci_slot)
        else:
            LOG.error(_LE("No eSwitch found for Fabric %s"), fabric)

        return pci_slot

    def delete_port(self, fabric, vnic_mac):
        dev = None
        eswitch = self._get_vswitch_for_fabric(fabric)
        if eswitch:
            dev = eswitch.detach_vnic(vnic_mac)
            if dev:
                self._config_vf_mac_address(fabric, dev)
        else:
            LOG.warning(_LW("No eSwitch found for Fabric %s"), fabric)
        return dev

    def port_release(self, fabric, vnic_mac):
        ret = None
        eswitch = self._get_vswitch_for_fabric(fabric)
        dev = eswitch.get_dev_for_vnic(vnic_mac)
        if dev:
            if (eswitch.get_port_state(dev) ==
                    constants.VPORT_STATE_UNPLUGGED):
                ret = self.set_vlan(
                    fabric, vnic_mac, constants.UNTAGGED_VLAN_ID)
                self.port_down(fabric, vnic_mac)
        eswitch = self._get_vswitch_for_fabric(fabric)
        eswitch.port_release(vnic_mac)
        return ret

    def port_up(self, fabric, vnic_mac):
        eswitch = self._get_vswitch_for_fabric(fabric)
        if eswitch:
            dev = eswitch.get_dev_for_vnic(vnic_mac)
            if not dev:
                LOG.info(_LI("No device for MAC %s"), vnic_mac)

    def port_down(self, fabric, vnic_mac):
        eswitch = self._get_vswitch_for_fabric(fabric)
        if eswitch:
            dev = eswitch.get_dev_for_vnic(vnic_mac)
            if dev:
                LOG.info(_LI("IB port for MAC %s doen't support "
                         "port down"), vnic_mac)
            else:
                LOG.info(_LI("No device for MAC %s"), vnic_mac)

    def set_vlan(self, fabric, vnic_mac, vlan):
        eswitch = self._get_vswitch_for_fabric(fabric)
        if eswitch:
            eswitch.set_vlan(vnic_mac, vlan)
            dev = eswitch.get_dev_for_vnic(vnic_mac)
            state = eswitch.get_port_state(dev)
            if dev:
                if state in (constants.VPORT_STATE_ATTACHED,
                             constants.VPORT_STATE_UNPLUGGED):
                    if eswitch.get_port_table()[dev]['alias']:
                        dev = eswitch.get_port_table()[dev]['alias']
                    try:
                        self._config_vlan_ib(fabric, dev, vlan)
                        return True
                    except RuntimeError:
                        LOG.error(_LE('Set VLAN operation failed'))
        return False

    def get_eswitch_tables(self, fabrics):
        tables = {}
        for fabric in fabrics:
            eswitch = self._get_vswitch_for_fabric(fabric)
            if eswitch:
                tables[fabric] = {
                    'port_table': eswitch.get_port_table_matrix(),
                    'port_policy': eswitch.get_port_policy_matrix()
                }
            else:
                LOG.info(_LI("Get eswitch tables: No eswitch %s") % fabric)
        return tables

    def _get_vswitch_for_fabric(self, fabric):
        if fabric in self.eswitches:
            return self.eswitches[fabric]
        else:
            return

    def _config_vf_pkey(self, ppkey_idx, pkey_idx,
                        pf_mlx_dev, vf_pci_id, hca_port):
        path = constants.PKEY_INDEX_PATH % (pf_mlx_dev, vf_pci_id,
                                            hca_port, pkey_idx)
        cmd = ['ebrctl', 'write-sys', path, ppkey_idx]
        execute(cmd, root_helper=None)

    def _get_guid_idx(self, pf_mlx_dev, dev, hca_port):
        path = constants.GUID_INDEX_PATH % (pf_mlx_dev, dev, hca_port)
        with open(path) as fd:
            idx = fd.readline().strip()
        return idx

    def _get_guid_from_mac(self, mac, device_type):
        guid = None
        if device_type == constants.CX3_VF_DEVICE_TYPE:
            if mac is None:
                guid = constants.INVALID_GUID_CX3
            else:
                mac = mac.replace(':', '')
                prefix = mac[:6]
                suffix = mac[6:]
                guid = prefix + '0000' + suffix
        elif device_type == constants.CX4_VF_DEVICE_TYPE:
            if mac is None:
                guid = constants.INVALID_GUID_CX4
            else:
                prefix = mac[:9]
                suffix = mac[9:]
                guid = prefix + '00:00:' + suffix
        return guid

    def _config_vf_mac_address(self, fabric, dev, vnic_mac=None):
        fabric_details = self.rm.get_fabric_details(fabric)
        vf_device_type = fabric_details['vfs'][dev]['vf_device_type']
        vguid = self._get_guid_from_mac(vnic_mac, vf_device_type)
        if vf_device_type == constants.CX3_VF_DEVICE_TYPE:
            self._config_vf_mac_address_cx3(vguid, dev, fabric_details)
        elif vf_device_type == constants.CX4_VF_DEVICE_TYPE:
            self._config_vf_mac_address_cx4(vguid, dev, fabric_details)
        else:
            LOG.error(_LE("Unsupported vf device type: %s "),
                      vf_device_type)

    def _config_vf_mac_address_cx3(self, vguid, dev, fabric_details):
        hca_port = fabric_details['hca_port']
        pf_mlx_dev = fabric_details['pf_mlx_dev']
        self._config_vf_pkey(
            INVALID_PKEY, DEFAULT_PKEY_IDX, pf_mlx_dev, dev, hca_port)

        guid_idx = self._get_guid_idx(pf_mlx_dev, dev, hca_port)
        path = constants.ADMIN_GUID_PATH % (pf_mlx_dev, hca_port, guid_idx)
        cmd = ['ebrctl', 'write-sys', path, vguid]
        execute(cmd, root_helper=None)
        ppkey_idx = self._get_pkey_idx(
            int(DEFAULT_PKEY, 16), pf_mlx_dev, hca_port)
        if ppkey_idx >= 0:
            self._config_vf_pkey(
                ppkey_idx, PARTIAL_PKEY_IDX, pf_mlx_dev, dev, hca_port)
        else:
            LOG.error(_LE("Can't find partial management pkey for"
                          "%(pf)s:%(dev)s"), {'pf': pf_mlx_dev, 'dev': dev})

    def _config_vf_mac_address_cx4(self, vguid, dev, fabric_details):
        vf_num = fabric_details['vfs'][dev]['vf_num']
        pf_mlx_dev = fabric_details['pf_mlx_dev']
        guid_node = constants.CX4_GUID_NODE_PATH % {'module': pf_mlx_dev,
                                                    'vf_num': vf_num}
        guid_port = constants.CX4_GUID_PORT_PATH % {'module': pf_mlx_dev,
                                                    'vf_num': vf_num}
        guid_poliy = constants.CX4_GUID_POLICY_PATH % {'module': pf_mlx_dev,
                                                       'vf_num': vf_num}
        for path in (guid_node, guid_port):
            cmd = ['ebrctl', 'write-sys', path, vguid]
            execute(cmd, root_helper=None)

        cmd = ['ebrctl', 'write-sys', guid_poliy, 'Up\n']
        execute(cmd, root_helper=None)

    def _config_vlan_ib(self, fabric, dev, vlan):
        fabric_details = self.rm.get_fabric_details(fabric)
        hca_port = fabric_details['hca_port']
        pf_mlx_dev = fabric_details['pf_mlx_dev']
        vf_device_type = fabric_details['vfs'][dev]['vf_device_type']
        if vf_device_type == constants.CX3_VF_DEVICE_TYPE:
            self._config_vlan_ib_cx3(vlan, pf_mlx_dev, dev, hca_port)
        elif vf_device_type == constants.CX4_VF_DEVICE_TYPE:
            pass
        else:
            LOG.error(_LE("Unsupported vf device type: %s "),
                      vf_device_type)

    def _config_vlan_ib_cx3(self, vlan, pf_mlx_dev, dev, hca_port):
        if vlan == 0:
            ppkey_idx = self._get_pkey_idx(
                int(DEFAULT_PKEY, 16), pf_mlx_dev, hca_port)
            if ppkey_idx >= 0:
                self._config_vf_pkey(
                    ppkey_idx, DEFAULT_PKEY_IDX, pf_mlx_dev, dev, hca_port)
        else:
            ppkey_idx = self._get_pkey_idx(str(vlan), pf_mlx_dev, hca_port)
            if ppkey_idx:
                self._config_vf_pkey(
                    ppkey_idx, DEFAULT_PKEY_IDX, pf_mlx_dev, dev, hca_port)

    def _get_pkey_idx(self, vlan, pf_mlx_dev, hca_port):
        PKEYS_PATH = "/sys/class/infiniband/%s/ports/%s/pkeys/*"
        paths = PKEYS_PATH % (pf_mlx_dev, hca_port)
        for path in glob.glob(paths):
            fd = open(path)
            pkey = fd.readline()
            fd.close()
            # the MSB in pkey is the membership bit ( 0 - partial, 1 - full)
            # the other 15 bit are the number of the pkey
            # so we want to remove the 16th bit when compare pkey file
            # to the vlan (pkey) we are looking for
            is_match = int(pkey, 16) & DEFAULT_MASK == int(vlan) & DEFAULT_MASK
            if is_match:
                return path.split('/')[-1]
        return None

    def _config_port_up(self, dev):
        cmd = ['ip', 'link', 'set', dev, 'up']
        execute(cmd, root_helper=None)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_mlnx._i18n import _LE, _LI
from oslo_log import log as logging

from networking_mlnx.eswitchd.common import constants

LOG = logging.getLogger(__name__)


class BasicMessageHandler(object):
    MSG_ATTRS_MANDATORY_MAP = set()

    def __init__(self, msg):
        self.msg = msg

    def execute(self):
        raise Exception("execute method MUST be implemented!")

    def validate(self):
        ret = True
        msg_attr = set(self.msg.keys())
        for attr in self.MSG_ATTRS_MANDATORY_MAP:
            if attr not in msg_attr:
                return False
        if 'vnic_type' in self.msg.keys():
            ret = self.validate_vnic_type(self.msg['vnic_type'])
        return ret

    def validate_vnic_type(self, vnic_type):
        if vnic_type in (constants.VIF_TYPE_HOSTDEV, ):
            return True
        return False

    def build_response(self, status, reason=None, response=None):
        if status:
            msg = {'status': 'OK', 'response': response}
        else:
            msg = {'status': 'FAIL', 'reason': reason}
        return msg


class AttachVnic(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = set(['fabric', 'vnic_type',
                                   'device_id', 'vnic_mac',
                                   'dev_name'])

    def __init__(self, msg):
        super(AttachVnic, self).__init__(msg)

    def execute(self, eswitch_handler):
        fabric = self.msg['fabric']
        vnic_type = self.msg['vnic_type']
        device_id = self.msg['device_id']
        vnic_mac = (self.msg['vnic_mac']).lower()
        dev_name = self.msg['dev_name']
        dev = eswitch_handler.create_port(
            fabric, vnic_type, device_id, vnic_mac, dev_name)
        if dev:
            return self.build_response(True, response={'dev': dev})
        else:
            return self.build_response(False, reason='Attach vnic failed')


class PlugVnic(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric', 'device_id',
                               'vnic_mac', 'dev_name')

    def __init__(self, msg):
        super(PlugVnic, self).__init__(msg)

    def execute(self, eswitch_handler):
        fabric = self.msg['fabric']
        device_id = self.msg['device_id']
        vnic_mac = (self.msg['vnic_mac']).lower()
        dev_name = self.msg['dev_name']

        dev = eswitch_handler.plug_nic(fabric, device_id, vnic_mac, dev_name)
        if dev:
            return self.build_response(True, response={'dev': dev})
        else:
            return self.build_response(False, reason='Plug vnic failed')


class DetachVnic(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric', 'vnic_mac')

    def __init__(self, msg):
        super(DetachVnic, self).__init__(msg)

    def execute(self, eswitch_handler):
        fabric = self.msg['fabric']
        vnic_mac = (self.msg['vnic_mac']).lower()
        dev = eswitch_handler.delete_port(fabric, vnic_mac)
        if dev:
            return self.build_response(True, response={'dev': dev})
        else:
            return self.build_response(True, response={})


class SetVLAN(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric', 'port_mac', 'vlan')

    def __init__(self, msg):
        super(SetVLAN, self).__init__(msg)

    def execute(self, eswitch_handler):
        fabric = self.msg['fabric']
        vnic_mac = (self.msg['port_mac']).lower()
        vlan = self.msg['vlan']
        ret = eswitch_handler.set_vlan(fabric, vnic_mac, vlan)
        reason = None
        if not ret:
            reason = 'Set VLAN Failed'
        if reason:
            return self.build_response(False, reason=reason)
        return self.build_response(True, response={})


class GetVnics(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric', )

    def __init__(self, msg):
        super(GetVnics, self).__init__(msg)

    def execute(self, eswitch_handler):
        fabric = self.msg['fabric']
        if fabric == '*':
            fabrics = eswitch_handler.eswitches.keys()
            LOG.info(_LI("fabrics = %s") % fabrics)
        else:
            fabrics = [fabric]
        vnics = eswitch_handler.get_vnics(fabrics)
        return self.build_response(True, response=vnics)


class PortRelease(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric', 'ref_by', 'mac')

    def __init__(self, msg):
        super(PortRelease, self).__init__(msg)

    def execute(self, eswitch_handler):
        ref_by_keys = ['mac_address']
        fabric = self.msg['fabric']
        vnic_mac = (self.msg['mac']).lower()
        ref_by = self.msg['ref_by']
        reason = None
        if ref_by not in ref_by_keys:
            reason = "reb_by %s is not supported" % ref_by
        else:
            try:
                eswitch_handler.port_release(fabric, vnic_mac)
            except Exception:
                reason = "port release failed"
                LOG.exception(_LE("PortRelease failed"))
        if reason:
            return self.build_response(False, reason=reason)
        return self.build_response(True, response={})


class SetFabricMapping(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric', 'interface')

    def __init__(self, msg):
        super(SetFabricMapping, self).__init__(msg)

    def execute(self, eswitch_handler):
        fabric = self.msg['fabric']
        interface = self.msg['interface']
        response = {'fabric': fabric, 'dev': interface}
        return self.build_response(True, response=response)


class PortUp(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric', 'mac')

    def __init__(self, msg):
        super(PortUp, self).__init__(msg)

    def execute(self, eswitch_handler):
        # fabric = self.msg['fabric']
        # mac = self.msg['mac']
        # eswitch_handler.port_up(fabric, mac)
        return self.build_response(True, response={})


class PortDown(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric', 'mac')

    def __init__(self, msg):
        super(PortDown, self).__init__(msg)

    def execute(self, eswitch_handler):
        # fabric = self.msg['fabric']
        # mac = self.msg['mac']
        # eswitch_handler.port_down(fabric, mac)
        return self.build_response(True, response={})


class GetEswitchTables(BasicMessageHandler):
    MSG_ATTRS_MANDATORY_MAP = ('fabric',)

    def __init__(self, msg):
        super(GetEswitchTables, self).__init__(msg)

    def execute(self, eswitch_handler):
        fabric = self.msg.get('fabric', '*')
        if fabric == '*':
            fabrics = eswitch_handler.eswitches.keys()
            LOG.info(_LI("fabrics = %s"), fabrics)
        else:
            fabrics = [fabric]
        response = {'tables': eswitch_handler.get_eswitch_tables(fabrics)}
        return self.build_response(True, response=response)


class MessageDispatch(object):
    MSG_MAP = {'create_port': AttachVnic,
               'delete_port': DetachVnic,
               'set_vlan': SetVLAN,
               'get_vnics': GetVnics,
               'port_release': PortRelease,
               'port_up': PortUp,
               'port_down': PortDown,
               'define_fabric_mapping': SetFabricMapping,
               'plug_nic': PlugVnic,
               'get_eswitch_tables': GetEswitchTables}

    def __init__(self, eswitch_handler):
        self.eswitch_handler = eswitch_handler

    def handle_msg(self, msg):
        LOG.info(_LI("Handling message - %s"), msg)
        result = {}
        action = msg.pop('action')

        if action in MessageDispatch.MSG_MAP.keys():
            msg_handler = MessageDispatch.MSG_MAP[action](msg)
            if msg_handler.validate():
                result = msg_handler.execute(self.eswitch_handler)
            else:
                LOG.error(_LE('Invalid message - cannot handle'))
                result = {'status': 'FAIL', 'reason': 'validation failed'}
        else:
            LOG.error(_LE("Unsupported action - %s"), action)
            result = {'action': action, 'status': 'FAIL',
                      'reason': 'unknown action'}
        result['action'] = action
        return result



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shlex
import subprocess

from networking_mlnx._i18n import _LI
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def get_root_helper():
    root_helper = 'sudo neutron-rootwrap %s' % cfg.CONF.DAEMON.rootwrap_conf
    return root_helper


def execute(cmd, root_helper=None, process_input=None, addl_env=None,
            check_exit_code=True, return_stderr=False):
    if not root_helper:
        root_helper = get_root_helper()
    cmd = shlex.split(root_helper) + cmd
    cmd = map(str, cmd)
    LOG.info(_LI("Running command: %s") % " ".join(cmd))
    env = os.environ.copy()
    if addl_env:
        env.update(addl_env)
    obj = subprocess.Popen(cmd, shell=False, stdin=subprocess.PIPE,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                           env=env)

    _stdout, _stderr = (process_input and
                        obj.communicate(process_input) or
                        obj.communicate())
    obj.stdin.close()
    m = ("\nCommand: %s\nExit code: %s\nStdout: %r\nStderr: %r" %
         (cmd, obj.returncode, _stdout, _stderr))
    LOG.info(m)
    if obj.returncode and check_exit_code:
        raise RuntimeError(m)
    return return_stderr and (_stdout, _stderr) or _stdout


def execute_bg(cmd, root_helper=None, log=None):
    if not root_helper:
        root_helper = get_root_helper()
    cmd = shlex.split(root_helper) + cmd
    cmd = map(str, cmd)
    LOG.info(_LI("Running command: %s") % " ".join(cmd))



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_log import log as logging
import six

LOG = logging.getLogger(__name__)


class DeviceDB(object):
    def __init__(self):
        self.device_db = {}

    def get_pf(self, fabric):
        return self.device_db[fabric]['pf']

    def add_fabric(self, fabric, pf, pci_id, hca_port, fabric_type,
                   pf_mlx_dev):
        details = {}
        details['vfs'] = {}
        details['pf'] = pf
        details['pf_device_type'] = None
        details['pci_id'] = pci_id
        details['hca_port'] = hca_port
        details['fabric_type'] = fabric_type
        details['pf_mlx_dev'] = pf_mlx_dev
        self.device_db[fabric] = details

    def get_fabric_details(self, fabric):
        return self.device_db[fabric]

    def set_fabric_devices(self, fabric, vfs):
        self.device_db[fabric]['vfs'] = vfs
        vf = six.next(six.itervalues(vfs))
        self.device_db[fabric]['pf_device_type'] = vf['vf_device_type']

    def get_dev_fabric(self, dev):
        for fabric in self.device_db:
            if dev in self.device_db[fabric]['vfs']:
                return fabric



# Copyright (c) 2014 OpenStack Foundation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.extensions import portbindings
from neutron.plugins.common import constants as p_constants
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import mech_agent

AGENT_TYPE_MLNX = 'Mellanox plugin agent'
VIF_TYPE_IB_HOSTDEV = 'ib_hostdev'


class MlnxMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """Attach to networks using Mellanox eSwitch L2 agent.

    The MellanoxMechanismDriver integrates the ml2 plugin with the
    Mellanox eswitch L2 agent. Port binding with this driver requires the
    Mellanox eswitch  agent to be running on the port's host, and that agent
    to have connectivity to at least one segment of the port's
    network.
    """

    def __init__(self):
        super(MlnxMechanismDriver, self).__init__(
            agent_type=AGENT_TYPE_MLNX,
            vif_type=VIF_TYPE_IB_HOSTDEV,
            vif_details={portbindings.CAP_PORT_FILTER: False},
            supported_vnic_types=[portbindings.VNIC_DIRECT])

    def get_allowed_network_types(self, agent=None):
        return [p_constants.TYPE_LOCAL, p_constants.TYPE_FLAT,
                p_constants.TYPE_VLAN]

    def get_mappings(self, agent):
        return agent['configurations'].get('interface_mappings', {})

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        if self.check_segment_for_agent(segment, agent):
            if (segment[api.NETWORK_TYPE] in
                    (p_constants.TYPE_FLAT, p_constants.TYPE_VLAN)):
                self.vif_details['physical_network'] = segment[
                    'physical_network']
            context.set_binding(segment[api.ID],
                                self.vif_type,
                                self.vif_details)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob
import os
import re
import sys

import ethtool
from networking_mlnx._i18n import _LE, _LW
from oslo_log import log as logging

from networking_mlnx.eswitchd.common import constants
from networking_mlnx.eswitchd.utils.command_utils import execute

LOG = logging.getLogger(__name__)


class pciUtils(object):

    ETH_PATH = "/sys/class/net/%(interface)s"
    ETH_DEV = ETH_PATH + "/device"
    ETH_PORT = ETH_PATH + "/dev_id"
    PF_MLX_DEV_PATH = "/sys/class/infiniband/*"
    VENDOR_PATH = ETH_DEV + '/vendor'
    DEVICE_TYPE_PATH = ETH_DEV + '/virtfn%(vf_num)s/device'
    _VIRTFN_RE = re.compile("virtfn(?P<vf_num>\d+)")
    VFS_PATH = ETH_DEV + "/virtfn*"

    def get_vfs_info(self, pf):
        vfs_info = {}
        try:
            dev_path = self.ETH_DEV % {'interface': pf}
            dev_info = os.listdir(dev_path)
            for dev_filename in dev_info:
                result = self._VIRTFN_RE.match(dev_filename)
                if result and result.group('vf_num'):
                    dev_file = os.path.join(dev_path, dev_filename)
                    vf_pci = os.readlink(dev_file).strip("./")
                    vf_num = result.group('vf_num')
                    vf_device_type = self.get_vf_device_type(pf, vf_num)
                    vfs_info[vf_pci] = {'vf_num': vf_num,
                                        'vf_device_type': vf_device_type}
        except Exception:
            LOG.error(_LE("PCI device %s not found"), pf)
        return vfs_info

    def get_dev_attr(self, attr_path):
        try:
            fd = open(attr_path)
            return fd.readline().strip()
        except IOError:
            return

    def verify_vendor_pf(self, pf, vendor_id=constants.VENDOR):
        vendor_path = pciUtils.VENDOR_PATH % {'interface': pf}
        if self.get_dev_attr(vendor_path) == vendor_id:
            return True
        else:
            return False

    def get_vf_device_type(self, pf, vf_num):
        device_vf_type = None
        device_type_file = pciUtils.DEVICE_TYPE_PATH % {'interface': pf,
                                                        'vf_num': vf_num}
        try:
            with open(device_type_file, 'r') as fd:
                device_type = fd.read()
                device_type = device_type.strip(os.linesep)
                if device_type in constants.CX3_VF_DEVICE_TYPE_LIST:
                    device_vf_type = constants.CX3_VF_DEVICE_TYPE
                elif device_type in constants.CX4_VF_DEVICE_TYPE_LIST:
                    device_vf_type = constants.CX4_VF_DEVICE_TYPE
                elif device_type in constants.CX5_VF_DEVICE_TYPE_LIST:
                    device_vf_type = constants.CX5_VF_DEVICE_TYPE
        except IOError:
            pass
        return device_vf_type

    def is_sriov_pf(self, pf):
        vfs_path = pciUtils.VFS_PATH % {'interface': pf}
        vfs = glob.glob(vfs_path)
        if vfs:
            return True
        else:
            return

    def get_interface_type(self, ifc):
        cmd = ['ip', '-o', 'link', 'show', 'dev', ifc]
        try:
            result = execute(cmd, root_helper=None)
        except Exception as e:
            LOG.warning(_LW("Failed to execute command %(cmd)s due to %(e)s"),
                    {"cmd": cmd, "e": e})
            raise
        if result.find('link/ether') != -1:
            return 'eth'
        elif result.find('link/infiniband') != -1:
            return 'ib'
        else:
            return None

    def is_ifc_module(self, ifc, fabric_type):
        modules = {'eth': 'mlx4_en', 'ib': 'ipoib'}
        if modules[fabric_type] in ethtool.get_module(ifc):
            return True

    def filter_ifcs_module(self, ifcs, fabric_type):
        return [ifc for ifc in ifcs if self.is_ifc_module(ifc, fabric_type)]

    def get_auto_pf(self, fabric_type):
        def log_error_and_exit(err_msg):
            LOG.error(err_msg)
            sys.exit(1)

        mlnx_pfs = [ifc for ifc in ethtool.get_devices()
                    if self.verify_vendor_pf(ifc)]
        if not mlnx_pfs:
            log_error_and_exit("Didn't find any Mellanox devices.")

        mlnx_pfs = [ifc for ifc in mlnx_pfs if self.is_sriov_pf(ifc)]
        if not mlnx_pfs:
            log_error_and_exit("Didn't find Mellanox NIC "
                               "with SR-IOV capabilities.")
        mlnx_pfs = self.filter_ifcs_module(mlnx_pfs, fabric_type)
        if not mlnx_pfs:
            log_error_and_exit("Didn't find Mellanox NIC of type %s with "
                               "SR-IOV capabilites." % fabric_type)
        if len(mlnx_pfs) != 1:
            log_error_and_exit("Found multiple PFs %s. Configure Manually."
                               % mlnx_pfs)
        return mlnx_pfs[0]

    def get_eth_vf(self, dev):
        vf_path = pciUtils.ETH_DEV % {'interface': dev}
        try:
            device = os.readlink(vf_path)
            vf = device.split('/')[3]
            return vf
        except Exception:
            return None

    def get_pf_pci(self, pf, type=None):
        vf = self.get_eth_vf(pf)
        if vf:
            if type == 'normal':
                return vf
            else:
                return vf[:-2]
        return None

    def get_pf_mlx_dev(self, pci_id):
        paths = glob.glob(pciUtils.PF_MLX_DEV_PATH)
        for path in paths:
            id = os.readlink(path).split('/')[5]
            if pci_id == id:
                return path.split('/')[-1]

    def get_guid_index(self, pf_mlx_dev, dev, hca_port):
        guid_index = None
        path = constants.GUID_INDEX_PATH % (pf_mlx_dev, dev, hca_port)
        with open(path) as fd:
            guid_index = fd.readline().strip()
        return guid_index

    def get_eth_port(self, dev):
        port_path = pciUtils.ETH_PORT % {'interface': dev}
        try:
            with open(port_path) as f:
                dev_id = int(f.read(), 0)
                return dev_id + 1
        except IOError:
            return

    def get_vfs_macs_ib(self, fabric_details):
        if fabric_details['pf_device_type'] == constants.CX3_VF_DEVICE_TYPE:
            return self.get_vfs_macs_ib_cx3(fabric_details)
        elif fabric_details['pf_device_type'] == constants.CX4_VF_DEVICE_TYPE:
            return self.get_vfs_macs_ib_cx4(fabric_details)

    def get_vfs_macs_ib_cx3(self, fabric_details):
        hca_port = fabric_details['hca_port']
        pf_mlx_dev = fabric_details['pf_mlx_dev']
        macs_map = {}
        guids_path = constants.ADMIN_GUID_PATH % (pf_mlx_dev, hca_port,
                                                  '[1-9]*')
        paths = glob.glob(guids_path)
        for path in paths:
            vf_index = path.split('/')[-1]
            with open(path) as f:
                guid = f.readline().strip()
                if guid == constants.INVALID_GUID_CX3:
                    mac = constants.INVALID_MAC
                else:
                    head = guid[:6]
                    tail = guid[-6:]
                    mac = ":".join(re.findall('..?', head + tail))
                macs_map[str(int(vf_index))] = mac
        return macs_map

    def get_vfs_macs_ib_cx4(self, fabric_details):
        vfs = fabric_details['vfs']
        macs_map = {}
        for vf in vfs.values():
            vf_num = vf['vf_num']
            pf_mlx_dev = fabric_details['pf_mlx_dev']
            guid_path = constants.CX4_GUID_NODE_PATH % {'module': pf_mlx_dev,
                                                        'vf_num': vf_num}
            with open(guid_path) as f:
                guid = f.readline().strip()
                head = guid[:8]
                tail = guid[-9:]
                mac = head + tail
            macs_map[vf_num] = mac
        return macs_map

    def get_device_address(self, hostdev):
        domain = hostdev.attrib['domain'][2:]
        bus = hostdev.attrib['bus'][2:]
        slot = hostdev.attrib['slot'][2:]
        function = hostdev.attrib['function'][2:]
        dev = "%.4s:%.2s:%2s.%.1s" % (domain, bus, slot, function)
        return dev



#!/usr/bin/env python
# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import mock
import subprocess
import sys

sys.modules['ethtool'] = mock.Mock()

from networking_mlnx._i18n import _LE
from networking_mlnx.eswitchd.utils import pci_utils
from networking_mlnx.eswitchd.utils.pci_utils import pciUtils
from networking_mlnx.tests import base


class TestPciUtils(base.TestCase):

    def setUp(self):
        super(TestPciUtils, self).setUp()
        self.pci_utils = pci_utils.pciUtils()

    def _assert_get_auto_pf_error(self, log_msg):
        with mock.patch.object(pci_utils, 'LOG') as LOG:
            self.assertRaises(SystemExit,
                              self.pci_utils.get_auto_pf, 'fabtype')
            LOG.error.assert_called_with(log_msg)

    def _test_get_auto_pf(self, devices=[], is_vendor_pf=True,
                          is_sriov=True, valid_fabric_type=True):
        ifcs = devices if valid_fabric_type else []
        return contextlib.nested(
            mock.patch('ethtool.get_devices', return_value=devices),
            mock.patch.object(self.pci_utils, 'verify_vendor_pf',
                              return_value=is_vendor_pf),
            mock.patch.object(self.pci_utils, 'is_sriov_pf',
                              return_value=is_sriov),
            mock.patch.object(self.pci_utils, 'filter_ifcs_module',
                              return_value=ifcs),
        )

    def test_get_auto_pf_no_mlnx_devices(self):
        log_msg = "Didn't find any Mellanox devices."
        with self._test_get_auto_pf():
            self._assert_get_auto_pf_error(log_msg)

        log_msg = "Didn't find any Mellanox devices."
        with self._test_get_auto_pf(devices=['device-1'], is_vendor_pf=False):
            self._assert_get_auto_pf_error(log_msg)

    def test_get_auto_pf_no_mlnx_sriov_devices(self):
        log_msg = "Didn't find Mellanox NIC with SR-IOV capabilities."
        with self._test_get_auto_pf(devices=['device-1'], is_sriov=False):
            self._assert_get_auto_pf_error(log_msg)

    def test_get_auto_pf_wrong_fabric_type(self):
        log_msg = ("Didn't find Mellanox NIC of type fabtype with "
                   "SR-IOV capabilites.")
        with self._test_get_auto_pf(devices=['device-1'],
                                    valid_fabric_type=False):
            self._assert_get_auto_pf_error(log_msg)

    def test_get_auto_pf_multiple_pfs(self):
        devices = ['device-1', 'device-2']
        log_msg = "Found multiple PFs %s. Configure Manually." % devices
        with self._test_get_auto_pf(devices=devices):
            self._assert_get_auto_pf_error(log_msg)

    def test_get_vfs_info_not_found_device(self):
        pf = "pf_that_does_not_exist"
        with mock.patch.object(pci_utils, 'LOG') as LOG:
            self.pci_utils.get_vfs_info(pf)
            LOG.error.assert_called_with(_LE("PCI device %s not found"), pf)

    def test_get_dev_attr_valid_attr(self):
        cmd = "find /sys/class/net/*/device/vendor | head -1 | cut -d '/' -f5"
        pf = subprocess.check_output(cmd, shell=True)
        pf = pf.strip()
        if pf:
            attr_path = "/sys/class/net/%s/device/vendor" % pf
            return_val = self.pci_utils.get_dev_attr(attr_path)
            self.assertIsNotNone(return_val)

    def test_get_dev_attr_invalid_attr(self):
        attr_path = "/path/that/does/not/exist"
        return_val = self.pci_utils.get_dev_attr(attr_path)
        self.assertIsNone(return_val)

    def test_verify_vendor_pf_valid_vendor(self):
        cmd = "find /sys/class/net/*/device/vendor | head -1 | cut -d '/' -f5"
        pf = subprocess.check_output(cmd, shell=True)
        pf = pf.strip()
        if pf:
            attr_path = "/sys/class/net/%s/device/vendor" % pf
            attr = subprocess.check_output("cat %s" % attr_path, shell=True)
            attr = attr.strip()
            return_val = self.pci_utils.verify_vendor_pf(pf, attr)
            self.assertTrue(return_val)

    def test_verify_vendor_pf_invalid_vendor(self):
        cmd = "ls -U /sys/class/net | head -1"
        pf = subprocess.check_output(cmd, shell=True)
        pf = pf.strip()
        attr = "0x0000"

        return_val = self.pci_utils.verify_vendor_pf(pf, attr)
        self.assertFalse(return_val)

    def test_is_sriov_pf_false(self):
        pf = "pf_that_does_not_exist"

        is_sriov = self.pci_utils.is_sriov_pf(pf)
        self.assertFalse(is_sriov)

    def test_get_eth_vf_valid(self):
        cmd = "ls -U /sys/class/net | head -1"
        pf = subprocess.check_output(cmd, shell=True)
        pf = pf.strip()

        ret_val = self.pci_utils.get_eth_vf(pf)
        cmd = "ls -l /sys/class/net/%s/device | awk '{print $NF}'" \
            " | cut -d '/' -f4" % pf
        pci = subprocess.check_output(cmd, shell=True)
        pci = pci.strip()

        self.assertEqual(ret_val, pci)

    def test_get_eth_vf_invalid(self):
        pf = "pf_that_does_not_exist"
        ret_val = self.pci_utils.get_eth_vf(pf)
        self.assertIsNone(ret_val)

    def test_get_pf_pci_type_normal(self):
        cmd = "ls -U /sys/class/net | head -1"
        pf = subprocess.check_output(cmd, shell=True)
        pf = pf.strip()

        ret_val = self.pci_utils.get_pf_pci(pf, "normal")
        cmd = "ls -l /sys/class/net/%s/device | awk '{print $NF}'" \
            " | cut -d '/' -f4" % pf
        pci = subprocess.check_output(cmd, shell=True)
        pci = pci.strip()

        self.assertEqual(ret_val, pci)

    @mock.patch('networking_mlnx.eswitchd.utils.pci_utils.os.readlink')
    @mock.patch('networking_mlnx.eswitchd.utils.pci_utils.os')
    def test_get_pf_pci_type_none(self, mock_os, mock_readlink):
        inst = pciUtils()
        mock_readlink.return_value = "../../../0000:81:00.0"
        assert inst.get_pf_pci("dev") == "0000:81:00"

    def test_get_pf_pci_none(self):
        pf = "pf_that_does_not_exist"
        ret_val = self.pci_utils.get_pf_pci(pf)
        self.assertIsNone(ret_val)



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

VENDOR = '0x15b3'
VIF_TYPE_HOSTDEV = 'ib_hostdev'


VPORT_STATE_ATTACHED = 'attached'
VPORT_STATE_PENDING = 'pending'
VPORT_STATE_UNPLUGGED = 'unplugged'

UNTAGGED_VLAN_ID = 4095

INVALID_MAC = '00:00:00:00:00:00'

# CX3
ADMIN_GUID_PATH = "/sys/class/infiniband/%s/iov/ports/%s/admin_guids/%s"
GUID_INDEX_PATH = "/sys/class/infiniband/%s/iov/%s/ports/%s/gid_idx/0"
PKEY_INDEX_PATH = "/sys/class/infiniband/%s/iov/%s/ports/%s/pkey_idx/%s"

# CX4
CX4_GUID_NODE_PATH = ('/sys/class/infiniband/%(module)s/device/sriov/'
                      '%(vf_num)s/node')
CX4_GUID_PORT_PATH = ('/sys/class/infiniband/%(module)s/device/sriov/'
                      '%(vf_num)s/port')
CX4_GUID_POLICY_PATH = ('/sys/class/infiniband/%(module)s/device/sriov/'
                        '%(vf_num)s/policy')


INVALID_GUID_CX3 = 'ffffffffffffffff'
INVALID_GUID_CX4 = 'ff:ff:ff:ff:ff:ff:ff:ff'

IFCS_PATH = '/sys/class/net/*'

CONN_URL = '%(transport)s://%(addr)s:%(port)s'

CX3_VF_DEVICE_TYPE_LIST = ('0x1004', )
CX4_VF_DEVICE_TYPE_LIST = ('0x1014', '0x1016')
CX5_VF_DEVICE_TYPE_LIST = ('0x1018', )

CX3_VF_DEVICE_TYPE = 'CX3'
CX4_VF_DEVICE_TYPE = 'CX4'
CX5_VF_DEVICE_TYPE = 'CX5'

SOCKET_OS_PORT = '60001'
SOCKET_OS_TRANSPORT = 'tcp'
SOCKET_OS_ADDR = '0.0.0.0'



#!/usr/bin/python
# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys

from networking_mlnx.eswitchd.cli import conn_utils
from networking_mlnx.eswitchd.cli import exceptions

client = conn_utils.ConnUtil()


def parse():
    """Main method that manages supported CLI commands.

    The actions that are supported throught the CLI are:
    write-sys, del-port, allocate-port and add-port
    Each action is matched with method that should handle it
    e.g. write-sys action is matched with  write_sys method
    """

    parser = argparse.ArgumentParser(prog='ebrctl')
    parser.add_argument('action', action='store_true')

    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('vnic_mac')
    parent_parser.add_argument('device_id')
    parent_parser.add_argument('fabric')
    parent_parser.add_argument('vnic_type')
    subparsers = parser.add_subparsers()

    parser_add_port = subparsers.add_parser('add-port',
                                            parents=[parent_parser])
    parser_add_port.add_argument('dev_name')
    parser_add_port.set_defaults(func=add_port)

    parser_add_port = subparsers.add_parser('allocate-port',
                                            parents=[parent_parser])
    parser_add_port.set_defaults(func=allocate_port)

    parser_del_port = subparsers.add_parser('del-port')
    parser_del_port.set_defaults(func=del_port)
    parser_del_port.add_argument('fabric')
    parser_del_port.add_argument('vnic_mac')

    parser_write_sys = subparsers.add_parser('write-sys')
    parser_write_sys.set_defaults(func=write_sys)
    parser_write_sys.add_argument('path')
    parser_write_sys.add_argument('value')

    args = parser.parse_args()
    args.func(args)


def allocate_port(args):
    try:
        dev = client.allocate_nic(args.vnic_mac, args.device_id,
                                  args.fabric, args.vnic_type)
    except exceptions.MlxException as e:
        sys.stderr.write("Error in allocate command")
        sys.stderr.write(e.message)
        sys.exit(1)
    sys.stdout.write(dev)
    sys.exit(0)


def add_port(args):
    try:
        dev = client.plug_nic(args.vnic_mac, args.device_id, args.fabric,
                              args.vnic_type, args.dev_name)

    except exceptions.MlxException as e:
        sys.stderr.write("Error in add-port command")
        sys.stderr.write(e.message)
        sys.exit(1)
    sys.stdout.write(dev)
    sys.exit(0)


def del_port(args):
    try:
        client.deallocate_nic(args.vnic_mac, args.fabric)
    except exceptions.MlxException as e:
        sys.stderr.write("Error in del-port command")
        sys.stderr.write(e.message)
        sys.exit(1)
    sys.exit(0)


def write_sys(args):
    try:
        fd = open(args.path, 'w')
        fd.write(args.value)
        fd.close()
    except Exception as e:
        sys.stderr.write("Error in write-sys command")
        sys.stderr.write(e.message)
        sys.exit(1)
    sys.exit(0)


def main():
    parse()



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_mlnx.eswitchd.common import constants
from oslo_serialization import jsonutils
import zmq

from networking_mlnx.eswitchd.cli import exceptions
from networking_mlnx.eswitchd.utils.helper_utils import set_conn_url

REQUEST_TIMEOUT = 50000


class ConnUtil(object):
    def __init__(self):
        self.__conn = None

        transport = constants.SOCKET_OS_TRANSPORT
        port = constants.SOCKET_OS_PORT
        addr = constants.SOCKET_OS_ADDR
        self.conn_url = set_conn_url(transport, addr, port)

    def send_msg(self, msg):
        context = zmq.Context()
        socket = context.socket(zmq.REQ)
        socket.setsockopt(zmq.LINGER, 0)
        socket.connect(self.conn_url)

        try:
            socket.send(msg)
            poller = zmq.Poller()
            poller.register(socket, zmq.POLLIN)
            conn = dict(poller.poll(REQUEST_TIMEOUT))
            if conn:
                if conn.get(socket) == zmq.POLLIN:
                    response_msg = socket.recv(zmq.NOBLOCK)
                    response = self.parse_response_msg(response_msg)
                    return response
            else:
                print ('no result received')
        finally:
            socket.close()
            context.term()

    def parse_response_msg(self, recv_msg):
        msg = jsonutils.loads(recv_msg)
        error_msg = " "
        if msg['status'] == 'OK':
            if 'response' in msg:
                return msg['response']
            return
        elif msg['status'] == 'FAIL':
            error_msg = "Action %s failed: %s" % (msg['action'], msg['reason'])
        else:
            error_msg = "Unknown operation status %s" % msg['status']
        raise exceptions.MlxException(error_msg)

    def allocate_nic(self, vnic_mac, device_id, fabric, vnic_type,
                     dev_name=None):
        msg = jsonutils.dumps({'action': 'create_port',
                          'vnic_mac': vnic_mac,
                          'device_id': device_id,
                          'fabric': fabric,
                          'vnic_type': vnic_type,
                          'dev_name': dev_name})
        recv_msg = self.send_msg(msg)
        try:
            dev = recv_msg['dev']
        except Exception:
            error_msg = "Failed to allocate %s on %s" % (vnic_mac, fabric)
            raise exceptions.MlxException(error_msg)
        return dev

    def plug_nic(self, vnic_mac, device_id, fabric, vif_type, dev_name):
        msg = jsonutils.dumps({'action': 'plug_nic',
                          'vnic_mac': vnic_mac,
                          'device_id': device_id,
                          'fabric': fabric,
                          'vnic_type': vif_type,
                          'dev_name': dev_name})

        recv_msg = self.send_msg(msg)
        try:
            dev = recv_msg['dev']
        except Exception:
            error_msg = "Failed to plug_nic %s on %s" % (vnic_mac, fabric)
            raise exceptions.MlxException(error_msg)
        return dev

    def deallocate_nic(self, vnic_mac, fabric):
        msg = jsonutils.dumps({'action': 'delete_port',
                          'fabric': fabric,
                          'vnic_mac': vnic_mac})
        recv_msg = self.send_msg(msg)
        try:
            dev = recv_msg['dev']
        except Exception:
            error_msg = "Failed to deallocate %s on %s" % (vnic_mac, fabric)
            raise exceptions.MlxException(error_msg)
        return dev

    def get_tables(self, fabric):
        msg = jsonutils.dumps({'action': 'get_eswitch_tables',
                          'fabric': fabric})
        recv_msg = self.send_msg(msg)
        tables = recv_msg['tables']
        return tables



# Copyright 2013 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from networking_mlnx._i18n import _LI
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)
DEFAULT_INTERFACE_MAPPINGS = []

mlx_daemon_opts = [
    cfg.StrOpt('socket_os_transport', default="tcp"),
    cfg.StrOpt('socket_os_port', default="60001"),
    cfg.StrOpt('socket_os_addr', default="0.0.0.0"),
    cfg.ListOpt('fabrics',
                default=DEFAULT_INTERFACE_MAPPINGS,
                help=("List of <physical_network>:<physical_interface>")),
    cfg.IntOpt('default_timeout',
               default=5000,
               help=('Default timeout waiting for messages')),
    cfg.IntOpt('max_polling_count',
               default=5,
               help=('Daemon will do sync after max_polling_count '
                     '* default_timeout')),
    cfg.StrOpt('rootwrap_conf',
               default='/etc/neutron/rootwrap.conf',
               help=('rootwrap configuration file'))
]


cfg.CONF.register_opts(mlx_daemon_opts, "DAEMON")
logging.register_options(cfg.CONF)


def init(args, **kwargs):
    cfg.CONF(args=args, project='eswitchd',
             **kwargs)


def setup_logging():
    """Sets up the logging options for a log with supplied name."""
    logging.setup(cfg.CONF, 'eswitchd')
    LOG.info(_LI("Logging enabled!"))
    LOG.info(_LI("%(prog)s Started!"),
             {'prog': sys.argv[0]})
    LOG.debug("command line: %s", " ".join(sys.argv))
